# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Support for using CDMRemote on a THREDDS Data Server (TDS)."""

from .dataset import Dataset

__all__ = ['Dataset']
